@@ -1,22 +1,27 @@
+from __future__ import annotations
+
 import asyncio
+from functools import partial
+from typing import Any, Callable
 
 from ..config import Config
-from ..typing import ASGIFramework
-from ..utils import invoke_asgi, LifespanFailure, LifespanTimeout
+from ..typing import AppWrapper, ASGIReceiveEvent, ASGISendEvent, LifespanScope
+from ..utils import LifespanFailureError, LifespanTimeoutError
 
 
-class UnexpectedMessage(Exception):
+class UnexpectedMessageError(Exception):
     pass
 
 
 class Lifespan:
-    def __init__(self, app: ASGIFramework, config: Config) -> None:
+    def __init__(self, app: AppWrapper, config: Config, loop: asyncio.AbstractEventLoop) -> None:
         self.app = app
         self.config = config
         self.startup = asyncio.Event()
         self.shutdown = asyncio.Event()
         self.app_queue: asyncio.Queue = asyncio.Queue(config.max_app_queue_size)
         self.supported = True
+        self.loop = loop
 
         # This mimics the Trio nursery.start task_status and is
         # required to ensure the support has been checked before
@@ -25,22 +30,38 @@ class Lifespan:
 
     async def handle_lifespan(self) -> None:
         self._started.set()
-        scope = {"type": "lifespan", "asgi": {"spec_version": "2.0"}}
+        scope: LifespanScope = {
+            "type": "lifespan",
+            "asgi": {"spec_version": "2.0", "version": "3.0"},
+        }
+
+        def _call_soon(func: Callable, *args: Any) -> Any:
+            future = asyncio.run_coroutine_threadsafe(func(*args), self.loop)
+            return future.result()
+
         try:
-            await invoke_asgi(self.app, scope, self.asgi_receive, self.asgi_send)
-        except LifespanFailure:
+            await self.app(
+                scope,
+                self.asgi_receive,
+                self.asgi_send,
+                partial(self.loop.run_in_executor, None),
+                _call_soon,
+            )
+        except LifespanFailureError:
             # Lifespan failures should crash the server
             raise
         except Exception:
             self.supported = False
             if not self.startup.is_set():
-                message = "ASGI Framework Lifespan error, continuing without Lifespan support"
+                await self.config.log.warning(
+                    "ASGI Framework Lifespan error, continuing without Lifespan support"
+                )
             elif not self.shutdown.is_set():
-                message = "ASGI Framework Lifespan error, shutdown without Lifespan support"
+                await self.config.log.exception(
+                    "ASGI Framework Lifespan error, shutdown without Lifespan support"
+                )
             else:
-                message = "ASGI Framework Lifespan errored after shutdown."
-
-            await self.config.log.exception(message)
+                await self.config.log.exception("ASGI Framework Lifespan errored after shutdown.")
         finally:
             self.startup.set()
             self.shutdown.set()
@@ -54,7 +75,7 @@ class Lifespan:
         try:
             await asyncio.wait_for(self.startup.wait(), timeout=self.config.startup_timeout)
         except asyncio.TimeoutError as error:
-            raise LifespanTimeout("startup") from error
+            raise LifespanTimeoutError("startup") from error
 
     async def wait_for_shutdown(self) -> None:
         await self._started.wait()
@@ -65,21 +86,21 @@ class Lifespan:
         try:
             await asyncio.wait_for(self.shutdown.wait(), timeout=self.config.shutdown_timeout)
         except asyncio.TimeoutError as error:
-            raise LifespanTimeout("shutdown") from error
+            raise LifespanTimeoutError("shutdown") from error
 
-    async def asgi_receive(self) -> dict:
+    async def asgi_receive(self) -> ASGIReceiveEvent:
         return await self.app_queue.get()
 
-    async def asgi_send(self, message: dict) -> None:
+    async def asgi_send(self, message: ASGISendEvent) -> None:
         if message["type"] == "lifespan.startup.complete":
             self.startup.set()
         elif message["type"] == "lifespan.shutdown.complete":
             self.shutdown.set()
         elif message["type"] == "lifespan.startup.failed":
             self.startup.set()
-            raise LifespanFailure("startup", message["message"])
+            raise LifespanFailureError("startup", message["message"])
         elif message["type"] == "lifespan.shutdown.failed":
             self.shutdown.set()
-            raise LifespanFailure("shutdown", message["message"])
+            raise LifespanFailureError("shutdown", message["message"])
         else:
-            raise UnexpectedMessage(message["type"])
+            raise UnexpectedMessageError(message["type"])
