@@ -1,3 +1,6 @@
+from __future__ import annotations
+
+import json
 import logging
 import os
 import sys
@@ -6,8 +9,11 @@ from http import HTTPStatus
 from logging.config import dictConfig, fileConfig
 from typing import Any, IO, Mapping, Optional, TYPE_CHECKING, Union
 
+import toml
+
 if TYPE_CHECKING:
     from .config import Config
+    from .typing import ResponseSummary, WWWScope
 
 
 def _create_logger(
@@ -24,7 +30,7 @@ def _create_logger(
     if target:
         logger = logging.getLogger(name)
         logger.handlers = [
-            logging.StreamHandler(sys_default) if target == "-" else logging.FileHandler(target)
+            logging.StreamHandler(sys_default) if target == "-" else logging.FileHandler(target)  # type: ignore # noqa: E501
         ]
         logger.propagate = propagate
         formatter = logging.Formatter(
@@ -55,16 +61,25 @@ class Logger:
         )
 
         if config.logconfig is not None:
-            log_config = {
-                "__file__": config.logconfig,
-                "here": os.path.dirname(config.logconfig),
-            }
-            fileConfig(config.logconfig, defaults=log_config, disable_existing_loggers=False)
+            if config.logconfig.startswith("json:"):
+                with open(config.logconfig[5:]) as file_:
+                    dictConfig(json.load(file_))
+            elif config.logconfig.startswith("toml:"):
+                with open(config.logconfig[5:]) as file_:
+                    dictConfig(toml.load(file_))
+            else:
+                log_config = {
+                    "__file__": config.logconfig,
+                    "here": os.path.dirname(config.logconfig),
+                }
+                fileConfig(config.logconfig, defaults=log_config, disable_existing_loggers=False)
         else:
             if config.logconfig_dict is not None:
                 dictConfig(config.logconfig_dict)
 
-    async def access(self, request: dict, response: dict, request_time: float) -> None:
+    async def access(
+        self, request: "WWWScope", response: "ResponseSummary", request_time: float
+    ) -> None:
         if self.access_logger is not None:
             self.access_logger.info(
                 self.access_log_format, self.atoms(request, response, request_time)
@@ -98,7 +113,9 @@ class Logger:
         if self.error_logger is not None:
             self.error_logger.log(level, message, *args, **kwargs)
 
-    def atoms(self, request: dict, response: dict, request_time: float) -> Mapping[str, str]:
+    def atoms(
+        self, request: "WWWScope", response: "ResponseSummary", request_time: float
+    ) -> Mapping[str, str]:
         """Create and return an access log atoms dictionary.
 
         This can be overidden and customised if desired. It should
@@ -111,7 +128,9 @@ class Logger:
 
 
 class AccessLogAtoms(dict):
-    def __init__(self, request: dict, response: dict, request_time: float) -> None:
+    def __init__(
+        self, request: "WWWScope", response: "ResponseSummary", request_time: float
+    ) -> None:
         for name, value in request["headers"]:
             self[f"{{{name.decode('latin1').lower()}}}i"] = value.decode("latin1")
         for name, value in response.get("headers", []):
@@ -128,9 +147,17 @@ class AccessLogAtoms(dict):
             remote_addr = client[0]
         else:  # make sure not to throw UnboundLocalError
             remote_addr = f"<???{client}???>"
-        method = request.get("method", "GET")
+        if request["type"] == "http":
+            method = request["method"]
+        else:
+            method = "GET"
         query_string = request["query_string"].decode()
         path_with_qs = request["path"] + ("?" + query_string if query_string else "")
+        status_code = response["status"]
+        try:
+            status_phrase = HTTPStatus(status_code).phrase
+        except ValueError:
+            status_phrase = f"<???{status_code}???>"
         self.update(
             {
                 "h": remote_addr,
@@ -139,7 +166,7 @@ class AccessLogAtoms(dict):
                 "r": f"{method} {request['path']} {protocol}",
                 "R": f"{method} {path_with_qs} {protocol}",
                 "s": response["status"],
-                "st": HTTPStatus(response["status"]).phrase,
+                "st": status_phrase,
                 "S": request["scheme"],
                 "m": method,
                 "U": request["path"],
