@@ -3,7 +3,7 @@ from __future__ import annotations
 from enum import auto, Enum
 from io import BytesIO, StringIO
 from time import time
-from typing import Awaitable, Callable, List, Optional, Tuple, Union
+from typing import Awaitable, Callable, Iterable, List, Optional, Tuple, Union
 from urllib.parse import unquote
 
 from wsproto.connection import Connection, ConnectionState, ConnectionType
@@ -23,13 +23,14 @@ from wsproto.utilities import generate_accept_token, split_comma_header
 from .events import Body, Data, EndBody, EndData, Event, Request, Response, StreamClosed
 from ..config import Config
 from ..typing import (
-    ASGIFramework,
+    AppWrapper,
     ASGISendEvent,
-    Context,
+    TaskGroup,
     WebsocketAcceptEvent,
     WebsocketResponseBodyEvent,
     WebsocketResponseStartEvent,
     WebsocketScope,
+    WorkerContext,
 )
 from ..utils import (
     build_and_validate_headers,
@@ -95,11 +96,13 @@ class Handshake:
         return True
 
     def accept(
-        self, subprotocol: Optional[str]
+        self,
+        subprotocol: Optional[str],
+        additional_headers: Iterable[Tuple[bytes, bytes]],
     ) -> Tuple[int, List[Tuple[bytes, bytes]], Connection]:
         headers = []
         if subprotocol is not None:
-            if subprotocol not in self.subprotocols:
+            if self.subprotocols is None or subprotocol not in self.subprotocols:
                 raise Exception("Invalid Subprotocol")
             else:
                 headers.append((b"sec-websocket-protocol", subprotocol.encode()))
@@ -120,6 +123,12 @@ class Handshake:
             headers.extend([(b"upgrade", b"WebSocket"), (b"connection", b"Upgrade")])
             status_code = 101
 
+        for name, value in additional_headers:
+            if b"sec-websocket-protocol" == name or name.startswith(b":"):
+                raise Exception(f"Invalid additional header, {name.decode()}")
+
+            headers.append((name, value))
+
         return status_code, headers, Connection(ConnectionType.SERVER, extensions)
 
 
@@ -154,9 +163,10 @@ class WebsocketBuffer:
 class WSStream:
     def __init__(
         self,
-        app: ASGIFramework,
+        app: AppWrapper,
         config: Config,
-        context: Context,
+        context: WorkerContext,
+        task_group: TaskGroup,
         ssl: bool,
         client: Optional[Tuple[str, int]],
         server: Optional[Tuple[str, int]],
@@ -170,6 +180,7 @@ class WSStream:
         self.closed = False
         self.config = config
         self.context = context
+        self.task_group = task_group
         self.response: WebsocketResponseStartEvent
         self.scope: WebsocketScope
         self.send = send
@@ -196,7 +207,7 @@ class WSStream:
             path, _, query_string = event.raw_path.partition(b"?")
             self.scope = {
                 "type": "websocket",
-                "asgi": {"spec_version": "2.1"},
+                "asgi": {"spec_version": "2.3", "version": "3.0"},
                 "scheme": self.scheme,
                 "http_version": event.http_version,
                 "path": unquote(path.decode("ascii")),
@@ -217,7 +228,7 @@ class WSStream:
                 await self._send_error_response(400)
                 self.closed = True
             else:
-                self.app_put = await self.context.spawn_app(
+                self.app_put = await self.task_group.spawn_app(
                     self.app, self.config, self.scope, self.app_send
                 )
                 await self.app_put({"type": "websocket.connect"})  # type: ignore
@@ -278,7 +289,10 @@ class WSStream:
             elif message["type"] == "websocket.close":
                 self.state = ASGIWebsocketState.CLOSED
                 await self._send_wsproto_event(
-                    CloseConnection(code=int(message.get("code", CloseReason.NORMAL_CLOSURE)))
+                    CloseConnection(
+                        code=int(message.get("code", CloseReason.NORMAL_CLOSURE)),
+                        reason=message.get("reason"),
+                    )
                 )
                 await self.send(EndData(stream_id=self.stream_id))
             else:
@@ -324,7 +338,9 @@ class WSStream:
 
     async def _accept(self, message: WebsocketAcceptEvent) -> None:
         self.state = ASGIWebsocketState.CONNECTED
-        status_code, headers, self.connection = self.handshake.accept(message.get("subprotocol"))
+        status_code, headers, self.connection = self.handshake.accept(
+            message.get("subprotocol"), message.get("headers", [])
+        )
         await self.send(
             Response(stream_id=self.stream_id, status_code=status_code, headers=headers)
         )
@@ -332,7 +348,7 @@ class WSStream:
             self.scope, {"status": status_code, "headers": []}, time() - self.start_time
         )
         if self.config.websocket_ping_interval is not None:
-            self.context.spawn(self._send_pings)
+            self.task_group.spawn(self._send_pings)
 
     async def _send_rejection(self, message: WebsocketResponseBodyEvent) -> None:
         body_suppressed = suppress_body("GET", self.response["status"])
