@@ -12,21 +12,18 @@ from .tcp_server import TCPServer
 from .udp_server import UDPServer
 from .worker_context import WorkerContext
 from ..config import Config, Sockets
-from ..typing import ASGIFramework
+from ..typing import AppWrapper
 from ..utils import (
     check_multiprocess_shutdown_event,
     load_application,
-    MustReloadError,
-    observe_changes,
     raise_shutdown,
     repr_socket_addr,
-    restart,
     ShutdownError,
 )
 
 
 async def worker_serve(
-    app: ASGIFramework,
+    app: AppWrapper,
     config: Config,
     *,
     sockets: Optional[Sockets] = None,
@@ -36,7 +33,6 @@ async def worker_serve(
     config.set_statsd_logger_class(StatsdLogger)
 
     lifespan = Lifespan(app, config)
-    reload_ = False
     context = WorkerContext()
 
     async with trio.open_nursery() as lifespan_nursery:
@@ -80,9 +76,6 @@ async def worker_serve(
             task_status.started(binds)
             try:
                 async with trio.open_nursery() as nursery:
-                    if config.use_reloader:
-                        nursery.start_soon(observe_changes, trio.sleep)
-
                     if shutdown_trigger is not None:
                         nursery.start_soon(raise_shutdown, shutdown_trigger)
 
@@ -96,22 +89,15 @@ async def worker_serve(
                     )
 
                     await trio.sleep_forever()
-            except trio.MultiError as error:
-                reload_ = any(isinstance(exc, MustReloadError) for exc in error.exceptions)
-            except MustReloadError:
-                reload_ = True
             except (ShutdownError, KeyboardInterrupt):
                 pass
             finally:
-                context.terminated = True
+                await context.terminated.set()
                 server_nursery.cancel_scope.deadline = trio.current_time() + config.graceful_timeout
 
         await lifespan.wait_for_shutdown()
         lifespan_nursery.cancel_scope.cancel()
 
-    if reload_:
-        restart()
-
 
 def trio_worker(
     config: Config, sockets: Optional[Sockets] = None, shutdown_event: Optional[EventType] = None
@@ -121,7 +107,7 @@ def trio_worker(
             sock.listen(config.backlog)
         for sock in sockets.insecure_sockets:
             sock.listen(config.backlog)
-    app = load_application(config.application_path)
+    app = load_application(config.application_path, config.wsgi_max_body_size)
 
     shutdown_trigger = None
     if shutdown_event is not None:
