@@ -1,41 +1,34 @@
-from typing import Any, Callable, Generator, Optional
+from __future__ import annotations
+
+from math import inf
+from typing import Any, Generator, Optional
 
 import trio
 
-from .context import Context
+from .task_group import TaskGroup
+from .worker_context import WorkerContext
 from ..config import Config
 from ..events import Closed, Event, RawData, Updated
 from ..protocol import ProtocolWrapper
-from ..typing import ASGIFramework
+from ..typing import AppWrapper
 from ..utils import parse_socket_addr
 
-MAX_RECV = 2 ** 16
-
-
-class EventWrapper:
-    def __init__(self) -> None:
-        self._event = trio.Event()
-
-    async def clear(self) -> None:
-        self._event = trio.Event()
-
-    async def wait(self) -> None:
-        await self._event.wait()
-
-    async def set(self) -> None:
-        self._event.set()
+MAX_RECV = 2**16
 
 
 class TCPServer:
-    def __init__(self, app: ASGIFramework, config: Config, stream: trio.abc.Stream) -> None:
+    def __init__(
+        self, app: AppWrapper, config: Config, context: WorkerContext, stream: trio.abc.Stream
+    ) -> None:
         self.app = app
         self.config = config
+        self.context = context
         self.protocol: ProtocolWrapper
         self.send_lock = trio.Lock()
-        self.timeout_lock = trio.Lock()
+        self.idle_lock = trio.Lock()
         self.stream = stream
 
-        self._keep_alive_timeout_handle: Optional[trio.CancelScope] = None
+        self._idle_handle: Optional[trio.CancelScope] = None
 
     def __await__(self) -> Generator[Any, None, None]:
         return self.run().__await__()
@@ -59,13 +52,13 @@ class TCPServer:
             client = parse_socket_addr(socket.family, socket.getpeername())
             server = parse_socket_addr(socket.family, socket.getsockname())
 
-            async with trio.open_nursery() as nursery:
-                self.nursery = nursery
-                context = Context(nursery)
+            async with TaskGroup() as task_group:
+                self._task_group = task_group
                 self.protocol = ProtocolWrapper(
                     self.app,
                     self.config,
-                    context,
+                    self.context,
+                    task_group,
                     ssl,
                     client,
                     server,
@@ -73,7 +66,7 @@ class TCPServer:
                     alpn_protocol,
                 )
                 await self.protocol.initiate()
-                await self._update_keep_alive_timeout()
+                await self._start_idle()
                 await self._read_data()
         except (trio.MultiError, OSError):
             pass
@@ -93,22 +86,23 @@ class TCPServer:
             await self._close()
             await self.protocol.handle(Closed())
         elif isinstance(event, Updated):
-            pass  # Triggers the keep alive timeout update
-        await self._update_keep_alive_timeout()
+            if event.idle:
+                await self._start_idle()
+            else:
+                await self._stop_idle()
 
     async def _read_data(self) -> None:
         while True:
             try:
-                data = await self.stream.receive_some(MAX_RECV)
+                with trio.fail_after(self.config.read_timeout or inf):
+                    data = await self.stream.receive_some(MAX_RECV)
             except (trio.ClosedResourceError, trio.BrokenResourceError):
                 await self.protocol.handle(Closed())
                 break
             else:
-                if data == b"":
-                    await self._update_keep_alive_timeout()
-                    break
                 await self.protocol.handle(RawData(data))
-                await self._update_keep_alive_timeout()
+                if data == b"":
+                    break
 
     async def _close(self) -> None:
         try:
@@ -124,29 +118,30 @@ class TCPServer:
             pass
         await self.stream.aclose()
 
-    async def _update_keep_alive_timeout(self) -> None:
-        async with self.timeout_lock:
-            if self._keep_alive_timeout_handle is not None:
-                self._keep_alive_timeout_handle.cancel()
-            self._keep_alive_timeout_handle = None
-            if self.protocol.idle:
-                self._keep_alive_timeout_handle = await self.nursery.start(
-                    _call_later, self.config.keep_alive_timeout, self._timeout
-                )
-
-    async def _timeout(self) -> None:
+    async def _initiate_server_close(self) -> None:
         await self.protocol.handle(Closed())
         await self.stream.aclose()
 
+    async def _start_idle(self) -> None:
+        async with self.idle_lock:
+            if self._idle_handle is None:
+                self._idle_handle = await self._task_group._nursery.start(self._run_idle)
 
-async def _call_later(
-    timeout: float,
-    callback: Callable,
-    task_status: trio._core._run._TaskStatus = trio.TASK_STATUS_IGNORED,
-) -> None:
-    cancel_scope = trio.CancelScope()
-    task_status.started(cancel_scope)
-    with cancel_scope:
-        await trio.sleep(timeout)
-        cancel_scope.shield = True
-        await callback()
+    async def _stop_idle(self) -> None:
+        async with self.idle_lock:
+            if self._idle_handle is not None:
+                self._idle_handle.cancel()
+            self._idle_handle = None
+
+    async def _run_idle(
+        self,
+        task_status: trio._core._run._TaskStatus = trio.TASK_STATUS_IGNORED,
+    ) -> None:
+        cancel_scope = trio.CancelScope()
+        task_status.started(cancel_scope)
+        with cancel_scope:
+            with trio.move_on_after(self.config.keep_alive_timeout):
+                await self.context.terminated.wait()
+
+            cancel_scope.shield = True
+            await self._initiate_server_close()
