@@ -1,8 +1,10 @@
+from __future__ import annotations
+
 import inspect
 import os
-import platform
 import socket
 import sys
+import time
 from enum import Enum
 from importlib import import_module
 from multiprocessing.synchronize import Event as EventType
@@ -20,33 +22,28 @@ from typing import (
     TYPE_CHECKING,
 )
 
+from .app_wrappers import ASGIWrapper, WSGIWrapper
 from .config import Config
-from .typing import (
-    ASGI2Framework,
-    ASGI3Framework,
-    ASGIFramework,
-    ASGIReceiveCallable,
-    ASGISendCallable,
-    Scope,
-)
+from .typing import AppWrapper, ASGIFramework, Framework, WSGIFramework
+
+try:
+    from typing import Literal
+except ImportError:
+    from typing_extensions import Literal  # type: ignore
 
 if TYPE_CHECKING:
     from .protocol.events import Request
 
 
-class Shutdown(Exception):
+class ShutdownError(Exception):
     pass
 
 
-class MustReloadException(Exception):
+class NoAppError(Exception):
     pass
 
 
-class NoAppException(Exception):
-    pass
-
-
-class LifespanTimeout(Exception):
+class LifespanTimeoutError(Exception):
     def __init__(self, stage: str) -> None:
         super().__init__(
             f"Timeout whilst awaiting {stage}. Your application may not support the ASGI Lifespan "
@@ -54,22 +51,22 @@ class LifespanTimeout(Exception):
         )
 
 
-class LifespanFailure(Exception):
+class LifespanFailureError(Exception):
     def __init__(self, stage: str, message: str) -> None:
         super().__init__(f"Lifespan failure in {stage}. '{message}'")
 
 
-class UnexpectedMessage(Exception):
+class UnexpectedMessageError(Exception):
     def __init__(self, state: Enum, message_type: str) -> None:
         super().__init__(f"Unexpected message type, {message_type} given the state {state}")
 
 
-class FrameTooLarge(Exception):
+class FrameTooLargeError(Exception):
     pass
 
 
 def suppress_body(method: str, status_code: int) -> bool:
-    return method == "HEAD" or 100 <= status_code < 200 or status_code in {204, 304, 412}
+    return method == "HEAD" or 100 <= status_code < 200 or status_code in {204, 304}
 
 
 def build_and_validate_headers(headers: Iterable[Tuple[bytes, bytes]]) -> List[Tuple[bytes, bytes]]:
@@ -78,27 +75,35 @@ def build_and_validate_headers(headers: Iterable[Tuple[bytes, bytes]]) -> List[T
     for name, value in headers:
         if name[0] == b":"[0]:
             raise ValueError("Pseudo headers are not valid")
-        validated_headers.append((bytes(name).lower().strip(), bytes(value).strip()))
+        validated_headers.append((bytes(name).strip(), bytes(value).strip()))
     return validated_headers
 
 
 def filter_pseudo_headers(headers: List[Tuple[bytes, bytes]]) -> List[Tuple[bytes, bytes]]:
     filtered_headers: List[Tuple[bytes, bytes]] = [(b"host", b"")]  # Placeholder
+    authority = None
+    host = b""
     for name, value in headers:
         if name == b":authority":  # h2 & h3 libraries validate this is present
-            filtered_headers[0] = (b"host", value)
+            authority = value
+        elif name == b"host":
+            host = value
         elif name[0] != b":"[0]:
             filtered_headers.append((name, value))
+    filtered_headers[0] = (b"host", authority if authority is not None else host)
     return filtered_headers
 
 
-def load_application(path: str) -> ASGIFramework:
-    try:
-        module_name, app_name = path.split(":", 1)
-    except ValueError:
+def load_application(path: str, wsgi_max_body_size: int) -> AppWrapper:
+    mode: Optional[Literal["asgi", "wsgi"]] = None
+    if ":" not in path:
         module_name, app_name = path, "app"
-    except AttributeError:
-        raise NoAppException()
+    elif path.count(":") == 2:
+        mode, module_name, app_name = path.split(":", 2)  # type: ignore
+        if mode not in {"asgi", "wsgi"}:
+            raise ValueError("Invalid mode, must be 'asgi', or 'wsgi'")
+    else:
+        module_name, app_name = path.split(":", 1)
 
     module_path = Path(module_name).resolve()
     sys.path.insert(0, str(module_path.parent))
@@ -110,17 +115,29 @@ def load_application(path: str) -> ASGIFramework:
         module = import_module(import_name)
     except ModuleNotFoundError as error:
         if error.name == import_name:
-            raise NoAppException()
+            raise NoAppError()
         else:
             raise
-
     try:
-        return eval(app_name, vars(module))
+        app = eval(app_name, vars(module))
     except NameError:
-        raise NoAppException()
+        raise NoAppError()
+    else:
+        return wrap_app(app, wsgi_max_body_size, mode)
 
 
-async def observe_changes(sleep: Callable[[float], Awaitable[Any]]) -> None:
+def wrap_app(
+    app: Framework, wsgi_max_body_size: int, mode: Optional[Literal["asgi", "wsgi"]]
+) -> AppWrapper:
+    if mode is None:
+        mode = "asgi" if is_asgi(app) else "wsgi"
+    if mode == "asgi":
+        return ASGIWrapper(cast(ASGIFramework, app))
+    else:
+        return WSGIWrapper(cast(WSGIFramework, app), wsgi_max_body_size)
+
+
+def wait_for_changes(shutdown_event: EventType) -> None:
     last_updates: Dict[Path, float] = {}
     for module in list(sys.modules.values()):
         filename = getattr(module, "__file__", None)
@@ -132,63 +149,28 @@ async def observe_changes(sleep: Callable[[float], Awaitable[Any]]) -> None:
         except (FileNotFoundError, NotADirectoryError):
             pass
 
-    while True:
-        await sleep(1)
+    while not shutdown_event.is_set():
+        time.sleep(1)
 
         for index, (path, last_mtime) in enumerate(last_updates.items()):
             if index % 10 == 0:
                 # Yield to the event loop
-                await sleep(0)
+                time.sleep(0)
 
             try:
                 mtime = path.stat().st_mtime
             except FileNotFoundError:
-                # File deleted
-                raise MustReloadException()
+                return
             else:
                 if mtime > last_mtime:
-                    raise MustReloadException()
+                    return
                 else:
                     last_updates[path] = mtime
 
 
-def restart() -> None:
-    # Restart  this process (only safe for dev/debug)
-    executable = sys.executable
-    script_path = Path(sys.argv[0]).resolve()
-    args = sys.argv[1:]
-    main_package = sys.modules["__main__"].__package__
-
-    if main_package is None:
-        # Executed by filename
-        if platform.system() == "Windows":
-            if not script_path.exists() and script_path.with_suffix(".exe").exists():
-                # quart run
-                executable = str(script_path.with_suffix(".exe"))
-            else:
-                # python run.py
-                args.append(str(script_path))
-        else:
-            if script_path.is_file() and os.access(script_path, os.X_OK):
-                # hypercorn run:app --reload
-                executable = str(script_path)
-            else:
-                # python run.py
-                args.append(str(script_path))
-    else:
-        # Executed as a module e.g. python -m run
-        module = script_path.stem
-        import_name = main_package
-        if module != "__main__":
-            import_name = f"{main_package}.{module}"
-        args[:0] = ["-m", import_name.lstrip(".")]
-
-    os.execv(executable, [executable] + args)
-
-
 async def raise_shutdown(shutdown_event: Callable[..., Awaitable[None]]) -> None:
     await shutdown_event()
-    raise Shutdown()
+    raise ShutdownError()
 
 
 async def check_multiprocess_shutdown_event(
@@ -225,30 +207,6 @@ def repr_socket_addr(family: int, address: tuple) -> str:
         return f"{address}"
 
 
-async def invoke_asgi(
-    app: ASGIFramework, scope: Scope, receive: ASGIReceiveCallable, send: ASGISendCallable
-) -> None:
-    if _is_asgi_2(app):
-        scope["asgi"]["version"] = "2.0"
-        app = cast(ASGI2Framework, app)
-        asgi_instance = app(scope)
-        await asgi_instance(receive, send)
-    else:
-        scope["asgi"]["version"] = "3.0"
-        app = cast(ASGI3Framework, app)
-        await app(scope, receive, send)
-
-
-def _is_asgi_2(app: ASGIFramework) -> bool:
-    if inspect.isclass(app):
-        return True
-
-    if hasattr(app, "__call__") and inspect.iscoroutinefunction(app.__call__):  # type: ignore
-        return False
-
-    return not inspect.iscoroutinefunction(app)
-
-
 def valid_server_name(config: Config, request: "Request") -> bool:
     if len(config.server_names) == 0:
         return True
@@ -259,3 +217,11 @@ def valid_server_name(config: Config, request: "Request") -> bool:
             host = value.decode()
             break
     return host in config.server_names
+
+
+def is_asgi(app: Any) -> bool:
+    if inspect.iscoroutinefunction(app):
+        return True
+    elif hasattr(app, "__call__"):
+        return inspect.iscoroutinefunction(app.__call__)
+    return False
